#include <cstdio>
#include <ctime>
#include <FL/fl_ask.H>
#include <FL/fl_file_chooser.H>         // FLTK file chooser
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Choice.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Return_Button.H>
#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "camera.h"
#include "light.h"
#include "pygmalion.h"
#include "init.h"
#include "operation.h"
#include "subdivide.h"
#include "fileio.h"
#include "glWindow.h"
#include "usersetting.h"

#ifdef __MINGW32__
#define isnan _isnan
#endif

#ifdef DEBUG
#define DebugOut(a); cout << a << endl;
#else
#define DebugOut(a);
#endif

char filename[1024] = ""; // needed in function Load(...)
char expFilename[1024] = ""; // export file name or temporary
int outputC=0; // output camera
int precision=6; // output precision

extern char defaultSub[255]; // default subdivide level

extern int vertexOffset;  
extern int edgeOffset;
extern int faceOffset;

////////////////////////////////////////////////////
// pygmalion.cc:geometry
extern vector<Vertex> vertex; // Vertex Vector
extern vector<Edge> edge; // Edge Vector
extern vector<Face> face; // Face Vector

extern vector<int> subdividedVertex; 
extern vector<int> subdividedEdge; // Edge
extern vector<int> subdividedFace; // Face
extern int subdivideLevel; // init value=0 mean no subdivision

// glWindow.cc:
extern vector<Camera> cameraList;
extern vector<Light> lightList;
// bounding box for clipping for drawing by OpenGL
extern double minX,minY,minZ;
extern double maxX,maxY,maxZ;

// pygmalion.cc:selection
extern vector<int> selectV;

extern enum {
  uGEOMETRY=1,
  uSELECTION=2,
  uCAMERA=4,
  uTEXTURE=8
}opCode;

extern int changed;

extern UserSetting* userSetting;

void
Load(char* name,int type){
  FILE* fp;
  char buf[8192];

  SetChanged(1);
  
  fp=fopen(name,"r");
  if(fp==NULL){
    fl_alert("Unable to open \'%s\' for reading!");
    return;
  }

  fscanf(fp,"%s ",buf); // header
  DebugOut("header checking...");
  if(strcmp("PYG",buf)){
    fl_alert("\'%s\' is not Pygmalion file!",name);
    return;
  }

  // needed offset by importing 
  vertexOffset=0;  
  edgeOffset=0;
  faceOffset=0;
  int tmpLevel;

  if(type==OPEN){ // Initialize
    strcpy(filename,name);
    PYGInit();
  }else { // type==Import
    vertexOffset=subdividedVertex[1];
    edgeOffset=subdividedEdge[1];
    faceOffset=subdividedFace[1];
    // Subdivide Level -> 0
    tmpLevel=subdivideLevel;
    for(int cnt1=0;cnt1<tmpLevel;cnt1++){
      Subdivide(-1);
    }
  }

  // version,comment
  fscanf(fp,"%s ",buf); // version
  DebugOut("version:" << buf);
  //fscanf(fp,"%s ",buf); // comment
  fgets(buf,254,fp); // comment
  DebugOut("comment==" << buf);

  int vertexSize,edgeSize,faceSize,cameraSize,lightSize;

  fscanf(fp,"%d %d %d %d %d",&vertexSize,&edgeSize,&faceSize,
	 &cameraSize,&lightSize);
  DebugOut("vSize,eSize,fSize,cSize,lSize" << vertexSize << ","
	   << edgeSize << "," << faceSize << "," << cameraSize << "," 
	   << lightSize);
  vertex.reserve(vertexSize+vertexOffset);
  edge.reserve(edgeSize+edgeOffset);
  face.reserve(faceSize+faceOffset);
  
  // camera
  for(int cnt1=0;cnt1<cameraSize;cnt1++){
    cameraList.push_back(Camera());
    cameraList.back().ReadFile(fp);
    CreateGLWUI();
  }
  // light
  for(int cnt1=0;cnt1<lightSize;cnt1++){
    lightList.push_back(Light());
    lightList.back().ReadFile(fp);    
  }
  // vertex
  for(int cnt1=0;cnt1<vertexSize;cnt1++){
    vertex.push_back(Vertex());
    vertex.back().ReadFile(fp);
  }
  // edge
  for(int cnt1=0;cnt1<edgeSize;cnt1++){
    edge.push_back(Edge());
    edge.back().ReadFile(fp);
  }
  // face
  for(int cnt1=0;cnt1<faceSize;cnt1++){
    face.push_back(Face());
    face.back().ReadFile(fp);
  }
  fclose(fp);
  SetChanged(0);

  subdividedVertex[1]=vertex.size();
  subdividedEdge[1]=edge.size();
  subdividedFace[1]=face.size();

  if(type==IMPORT){
    for(int cnt1=0;cnt1<tmpLevel;cnt1++){
      Subdivide(1);
    }
    for(int cnt1=vertexOffset;cnt1<subdividedVertex[1];cnt1++){
      selectV.push_back(cnt1);
    }
    RefineTotalV();
  }
  
  for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
    CheckBB(vertex[cnt1]);
  }
  
  //AlignViews();
  Fl::redraw();
  PushOP(uGEOMETRY|uSELECTION);
  DebugOut("all done.");
}

void
Save(char* name){
  FILE* fp;
  fp=fopen(name,"w");
  if(fp==NULL){
    fl_alert("Unable to create \'%s\' for writing!");
    //cout << "couldn't open." << endl;
    return;
  }
  strcpy(filename,name);

  fprintf(fp,"PYG\n");
  fprintf(fp,"version1.0\n"); // version1.0
  fprintf(fp,"# comment line \n");
  fprintf(fp,"%d %d %d %d %d\n",subdividedVertex[1],subdividedEdge[1],
	  subdividedFace[1],(int)cameraList.size(),(int)lightList.size());
  // camera
  for(int cnt1=0;cnt1<cameraList.size();cnt1++){
    cameraList[cnt1].WriteFile(fp);
  }
  // light
  for(int cnt1=0;cnt1<lightList.size();cnt1++){
    lightList[cnt1].WriteFile(fp);
  }
  // vertex
  for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
    vertex[cnt1].WriteFile(fp);
  }
  // edge
  for(int cnt1=0;cnt1<subdividedEdge[1];cnt1++){
    edge[cnt1].WriteFile(fp);
  }
  // face
  for(int cnt1=0;cnt1<subdividedFace[1];cnt1++){
    face[cnt1].WriteFile(fp);
  }
  fclose(fp);
  changed=1;
  SetChanged(0);
}

void
ImportPYG(){
  DebugOut("Import PYG");
  char *newfile;  

  newfile = fl_file_chooser("Import File?", "*.pyg", filename);
  if (newfile != NULL) Load(newfile,IMPORT);
}


void
ExportPOVCore(int level,int povFlag){
  char tmp[255];
  int cnt1;

  time_t startTime;
  time_t endTime;
  double workTime;
  // Time 
  startTime=time(NULL);

  DebugOut("In ExportPOVCore;filename==" << filename);
  if(!strcmp(filename,"")){
    sprintf(expFilename,"Untitled.pov");
  }else{    
    DebugOut("strlen==" << strlen(filename));
    for(cnt1=0;cnt1<strlen(filename)-4;cnt1++){
      tmp[cnt1]=filename[cnt1];
    }
    tmp[cnt1]='\0';
    DebugOut("tmp==" << tmp);
    sprintf(expFilename,"%s.pov",tmp);
  }
  // temporary export for rendering
  if(povFlag==TEMP_EXPORT){
    sprintf(expFilename,"%s",tmpExpFilename);
  }
  DebugOut("Export POV-Ray core");
  DebugOut("povFile==" << expFilename);
  

  FILE *fp,*lightFP,*cameraFP,*textureFP;
  int flag;
  fp=fopen(expFilename,"r"); 
  if(fp!=NULL && povFlag!=TEMP_EXPORT){
    char mes[255];
    sprintf(mes,"%s already exist. Overwrite?",expFilename);
    flag=fl_choice(mes,"yes","no",NULL);
    if(flag==1){ // no
      DebugOut("In ExportPOV;return");
      return;
    }else { // 2==yes
      fclose(fp);
      fp=fopen(expFilename,"w");
      DebugOut("Willing write...");
    }
  }else{
    fp=fopen(expFilename,"w");
    DebugOut("Willing write also...");
  }
  
  int dif=level-subdivideLevel;
  for(cnt1=0;cnt1<dif;cnt1++){
    Subdivide(1);
  }


  fprintf(fp,"// Generated by Pygmalion\n\n"); 
  fprintf(fp, "camera {\n");
  if(cameraList[outputC].type==PERSPECTIVE){
    fprintf(fp,"\tperspective\n");
    fprintf(fp,"\tlocation <%f, %f, %f>\n",cameraList[outputC].camera.X(),
	    cameraList[outputC].camera.Y(),-cameraList[outputC].camera.Z());
    fprintf(fp,"\tright %d*x\n",cameraList[outputC].width);
    fprintf(fp,"\tup %d*y\n",cameraList[outputC].height);
    fprintf(fp,"\tlook_at <%f, %f, %f>\n",cameraList[outputC].view.X(),
	    cameraList[outputC].view.Y(),-cameraList[outputC].view.Z());
    fprintf(fp,"\tangle %f\n",OGLAng2POVAng(cameraList[outputC],
					    cameraList[outputC].eyeAngle));
  }else{ // ORTHO
    float cameraCoord1=cameraList[outputC].camera.X();
    float cameraCoord2=cameraList[outputC].camera.Y();
    float cameraCoord3=-cameraList[outputC].camera.Z(); // left
    int signX,signY,signZ;
    switch(cameraList[outputC].type){
    case XYMP:
      signX=0;signY=0;signZ=1;
      cameraCoord3=-maxZ; // left coordinate
      break;
    case XYPM:
      signX=0;signY=0;signZ=-1;
      cameraCoord3=-minZ; // left 
      break;
    case YZMP:
      signX=+1;signY=0;signZ=0;
      cameraCoord1=minX;
      break;
    case YZPM:
      signX=-1;signY=0;signZ=0;
      cameraCoord1=maxX;
      break;
    case XZMP:
      signX=0;signY=+1;signZ=0;
      cameraCoord2=minY;
      break;
    case XZPM:
      signX=0;signY=-1;signZ=0;
      cameraCoord2=maxY;
      break;
    }
    fprintf(fp,"\torthographic\n");
    fprintf(fp, "\tlocation <%f, %f, %f>\n",
	    cameraCoord1,cameraCoord2,cameraCoord3);
    if(cameraList[outputC].width>cameraList[outputC].height){
      fprintf(fp,"\tright <2.0*%d/%d/%f, 0, 0>\n",
	      cameraList[outputC].width,
	      cameraList[outputC].height,
	      cameraList[outputC].zoom);
      fprintf(fp,"\tup <0, 2.0/%f, 0>\n",cameraList[outputC].zoom);
    }else{
      fprintf(fp,"\tright <2.0/%f, 0, 0>\n",cameraList[outputC].zoom);
      fprintf(fp,"\tup <0, 2.0*%d/%d/%f, 0>\n",
	      cameraList[outputC].height,
	      cameraList[outputC].width,
	      cameraList[outputC].zoom);
    }
    fprintf(fp,"\tdirection <0, 0, 1>\n");
    fprintf(fp,"\tlook_at <%f, %f, %f>\n",
	    cameraCoord1+signX,cameraCoord2+signY,cameraCoord3+signZ);
  }
  fprintf(fp, "}\n\n");

  Vertex* nowVertexP;
  Vector3d n0,n1,n2,n3; // vertex normals of a face
  // Face
  switch(povFlag){
  case TRIANGLE: // TRIANGLE
    fprintf(fp, "light_source { < 0, 1000, -2000 > color rgb <1, 1, 1> }\n");
    fprintf(fp, "light_source { < 2000, 100, 0 > color rgb <1, 1, 1> }\n");
    fprintf(fp, "light_source { < -500, -400, 2000 > color rgb <1, 1, 1> }\n\n");    
    fprintf(fp,"\n\n");
    fprintf(fp,"mesh {\n");

    for(cnt1=subdividedFace[level];cnt1<subdividedFace[level+1];cnt1++){
      if(face[cnt1].vertexLink.size()==3){
	fprintf(fp,"\ttriangle {\n"); 
	fprintf(fp,"\t\t<%.*f, %.*f, %.*f>,\n",
		precision,
		vertex[face[cnt1].vertexLink[0]].X(),
		precision,
		vertex[face[cnt1].vertexLink[0]].Y(),
		precision,
		-vertex[face[cnt1].vertexLink[0]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[1]].X(),
		vertex[face[cnt1].vertexLink[1]].Y(),
		-vertex[face[cnt1].vertexLink[1]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>}\n",
		vertex[face[cnt1].vertexLink[2]].X(),
		vertex[face[cnt1].vertexLink[2]].Y(),
		-vertex[face[cnt1].vertexLink[2]].Z());
      }else{
	fprintf(fp,"\ttriangle {\n"); 
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[0]].X(),
		vertex[face[cnt1].vertexLink[0]].Y(),
		-vertex[face[cnt1].vertexLink[0]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[1]].X(),
		vertex[face[cnt1].vertexLink[1]].Y(),
		-vertex[face[cnt1].vertexLink[1]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>}\n",
		vertex[face[cnt1].vertexLink[2]].X(),
		vertex[face[cnt1].vertexLink[2]].Y(),
		-vertex[face[cnt1].vertexLink[2]].Z());
	fprintf(fp,"\ttriangle {\n"); 
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[2]].X(),
		vertex[face[cnt1].vertexLink[2]].Y(),
		-vertex[face[cnt1].vertexLink[2]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[3]].X(),
		vertex[face[cnt1].vertexLink[3]].Y(),
		-vertex[face[cnt1].vertexLink[3]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>}\n",
		vertex[face[cnt1].vertexLink[0]].X(),
		vertex[face[cnt1].vertexLink[0]].Y(),
		-vertex[face[cnt1].vertexLink[0]].Z());
      }
    }
    break;

  case SMOOTH_TRIANGLE:    
  case TEMP_EXPORT:
    //vector<int> flag[vertex.size()]; // calculation flag

    fprintf(fp, "light_source { < %f, %f, %f > color rgb <1, 1, 1> }\n",
	    (cameraList[outputC].camera-cameraList[outputC].view).X()*1000
	    +cameraList[outputC].camera.X(),
	    (cameraList[outputC].camera-cameraList[outputC].view).Y()*1000
	    +cameraList[outputC].camera.Y(),
	    -(cameraList[outputC].camera-cameraList[outputC].view).Z()*1000
	    -cameraList[outputC].camera.Z());
    fprintf(fp,"\nmesh {\n");

    DebugOut("smooth_triangle");
    for(int nowFace=subdividedFace[level];nowFace<subdividedFace[level+1];
	nowFace++){// start face
      if(face[nowFace].vertexLink.size()==4){ // case: 4 vertices
	DebugOut("4 vertices face");
	DebugOut("nowFace==" << nowFace);

	for(int cntTmp=0;cntTmp<6;cntTmp++){ //start vertex
	  int cnt2=cntTmp;
	  if(cnt2>2){ // for second triangle
	    cnt2-=1;
	  }// cnt2 mean triangle vertex
	  int nowVertex=face[nowFace].vertexLink[cnt2%4];
	  DebugOut("nowVertex==" << nowVertex);
	  nowVertexP=&vertex[nowVertex];
	  vector<int> startEdge; // start edge number
	  vector<int> checkCnt; 

	  // find start 2 edges
	  int tmpNowV,opV,nowEdge;
	  for(int cnt3=0;cnt3<face[nowFace].vertexLink.size();cnt3++){
	    tmpNowV=face[nowFace].vertexLink[cnt3];
	    opV=face[nowFace].vertexLink[(cnt3+1)%4]; // opposite vertex
	    nowEdge=face[nowFace].FindEdge(tmpNowV,opV);
	    if(!vertex[nowVertex].HaveEdge(nowEdge)){
	      continue;
	    }
	    startEdge.push_back(nowEdge);
	    checkCnt.push_back(cnt3);
	  }
	  if(checkCnt[0]==0 && checkCnt[1]==3){
	    DebugOut("swap startEdge");
	    swap(startEdge[0],startEdge[1]);
	  }
#ifdef DEBUG
	  cout << "startEdge.size()==" << startEdge.size() << endl;
	  for(int tmp=0;tmp<startEdge.size();tmp++){
	    cout << "startEdge[" << tmp << "]==" << startEdge[tmp] << endl;
	  }
#endif
	  Vector3d totalNormal;
	  int faceSum=0;
	  Vector3d vec[3];
	  Vector3d tmpN;
	  vec[0]=vertex[nowVertex]-vertex[edge[startEdge[0]].
					 AnotherVertex(nowVertex)];
	  vec[1]=vertex[nowVertex]-vertex[edge[startEdge[1]].
					 AnotherVertex(nowVertex)];
	  
	  // can optimise.HasBreak...nowVertex-->cntTmp
	  if(face[nowFace].HasBreak(nowVertex)){ 
	  //if(cntTmp!=1 && cntTmp!=3){ //comment by Hidekuni Annaka 2000/10/20
	    vec[2]=vertex[nowVertex]
	      -vertex[face[nowFace].OppositeVertex(nowVertex)];
	    Vector3d tmpVec;
	    tmpVec.Cross(vec[0],vec[2]);
	    tmpVec.Normalize();
#ifdef DEBUG
	    tmpVec.Print();
#endif
	    if(isnan(tmpVec.coord[0])){
	      DebugOut("prevent nan : set zero");
	      tmpVec.Set(0,0,0); // prevent nan
	    }
	    tmpN.Cross(vec[2],vec[1]);
	    tmpN.Normalize();
#ifdef DEBUG
	    tmpN.Print();
#endif
	    if(isnan(tmpN.coord[0])){
	      DebugOut("prevent nan : set zero2");
	      tmpN.Set(0,0,0); // prevent nan
	    }
	    tmpN.Plus(tmpVec);
	    tmpN.Normalize();
	  }else{
	    tmpN.Cross(vec[0],vec[1]);
	    tmpN.Normalize(); // finish this face normal
	  }
	  if(isnan(tmpN.coord[0])){
	    DebugOut("prevent nan:abort this face2 -> set dummy");
	    tmpN.Set(0,0,0);
	  }
	  
#ifdef DEBUG
	  cout << "base normal==";
	  tmpN.Print();
#endif
	  totalNormal.Plus(tmpN); // add this face's normal
	  faceSum++;
	  for(int cnt3=0;cnt3<startEdge.size();cnt3++){
	    DebugOut("searching next face from edge[" << startEdge[cnt3] << "]...");
	    int oldEdge=startEdge[cnt3],newEdge;
	    int oldFace=nowFace,newFace;	    
	   
	    vec[0]=vertex[nowVertex]-vertex[edge[startEdge[0]].
					   AnotherVertex(nowVertex)];
	    vec[1]=vertex[nowVertex]-vertex[edge[startEdge[1]].
					   AnotherVertex(nowVertex)];
	    if(cnt3!=0){
	      swap(vec[0],vec[1]);
	    }
	    int whileFlag=1;
	    while(edge[oldEdge].facePLink.size()==2 && whileFlag){
	      DebugOut("found new face.");
	      newFace=edge[oldEdge].AnotherFace(oldFace); // newFace
	      DebugOut("oldEdge,oldFace,newFace==" << oldEdge << ", "
		       << oldFace << ", " << newFace);
	      if(newFace==nowFace){ // prevent loop
		DebugOut("but newFace is nowFace. so end nowFace.");
		cnt3=startEdge.size();
		whileFlag=0;
		continue;
	      }
	      vec[1]=vec[0];

	      int tmpEdge;
	      for(int cnt4=0;cnt4<nowVertexP->edgePLink.size();cnt4++){
		tmpEdge=nowVertexP->edgePLink[cnt4];
		if(face[newFace].HaveEdge(tmpEdge) && tmpEdge!=oldEdge){
		  newEdge=tmpEdge; // newEdge
		  cnt4=nowVertexP->edgePLink.size(); // exit this for-loop
		}
	      }
	      DebugOut("newEdge==" << newEdge);
	      vec[0]=vertex[nowVertex]
		-vertex[edge[newEdge].AnotherVertex(nowVertex)]; // another vector
	      if(face[newFace].HasBreak(nowVertex)){
		DebugOut("nowVertex in newFace has break");
		vec[2]=vertex[nowVertex]
		  -vertex[face[newFace].OppositeVertex(nowVertex)];
		Vector3d tmpVec;
		tmpVec.Cross(vec[0],vec[2]);
		tmpVec.Normalize();
		if(isnan(tmpVec.coord[0])){
		  DebugOut("prevent nan : through this face");
		  tmpVec.Set(0,0,0); // prevent nan
		}
		tmpN.Cross(vec[2],vec[1]);
		tmpN.Normalize();
		if(isnan(tmpN.coord[0])){
		  DebugOut("prevent nan : through this face2");
		  tmpN.Set(0,0,0); // prevent nan
		}
		tmpN.Plus(tmpVec);
		tmpN.Normalize();
	      }else{
		tmpN.Cross(vec[0],vec[1]);
		tmpN.Normalize();
	      }
	      if(isnan(tmpN.coord[0])){
		DebugOut("prevent NAN : set zero");
		tmpN.Set(0,0,0);
		faceSum--;
	      }
#ifdef DEBUG
	      cout << "tmpN==";
	      tmpN.Print();
#endif
	      if(cnt3==0){
		totalNormal.Plus(tmpN); // add this face's normal
	      }else{
		totalNormal.Minus(tmpN); // add this face's normal
	      }
	      faceSum++;
#ifdef DEBUG
	      cout << "now totalNormal==";
	      totalNormal.Print();
#endif
	      oldFace=newFace; // store face number
	      oldEdge=newEdge; // store edge number
	    }
	    DebugOut("end search.");
	  }
	  totalNormal.Normalize();
	  if(isnan(totalNormal.coord[0])){
	    DebugOut("set totalNormal 0,0,0");
	    totalNormal.Set(0,0,0);
	  }
#ifdef DEBUG
	  printf("\tfaceSum==%d\n",faceSum);
	  printf("\tAverage normal==");
	  totalNormal.Print();
#endif
	  // print POV smooth_triangle
	  if(cntTmp==0 || cntTmp==3){ // first vertex of triangle
	    fprintf(fp,"\tsmooth_triangle {\n");
	  }
	  fprintf(fp,"\t\t<%.*f, %.*f, %.*f>,<%f, %f, %f>",
		  precision,
		  vertex[nowVertex].X(),
		  precision,
		  vertex[nowVertex].Y(),
		  precision,
		  -vertex[nowVertex].Z(),
		  totalNormal.X(),
		  totalNormal.Y(),
		  -totalNormal.Z()); 	  
	  if(cntTmp==2 || cntTmp==5){ // first vertex of triangle
	    fprintf(fp,"}\n");
	  }else{
	    fprintf(fp,",\n");
	  }
	}

      }else{ // 3 vertices face
	DebugOut("3 vertices face");
	DebugOut("nowFace==" << nowFace);
	fprintf(fp,"\tsmooth_triangle {\n");
	for(int cnt2=0;cnt2<face[nowFace].vertexLink.size();
	    cnt2++){ //start vertex
	  int nowVertex=face[nowFace].vertexLink[cnt2];
	  DebugOut("nowVertex==" << nowVertex);
	  nowVertexP=&vertex[nowVertex];
	  vector<int> startEdge; // start edge number

	  int flag=0; 
	  int nowEdge;
	  for(int cnt3=0;cnt3<face[nowFace].edgeLink.size();cnt3++){
	    nowEdge=face[nowFace].edgeLink[cnt3];
	    if(!vertex[nowVertex].HaveEdge(nowEdge)){
	      continue;
	    }
	    startEdge.push_back(nowEdge);
	    flag+=cnt3;
	  }	  
	  if(flag==2){ // edgeLink[0] and edgeLink[2]
	    swap(startEdge[0],startEdge[1]); // sort startEdge in a face
	  }
#ifdef DEBUG
	  cout << "startEdge.size()==" << startEdge.size() << endl;
	  for(int tmp=0;tmp<startEdge.size();tmp++){
	    cout << "startEdge[" << tmp << "]==" << startEdge[tmp] << endl;
	  }
#endif
	  Vector3d totalNormal;
	  int faceSum=0;
	  Vector3d vec[3];
	  Vector3d tmpN,tmpN2;
	  //Vector3d tmpVec;
	  vec[0]=vertex[nowVertex]-vertex[edge[startEdge[0]].
					 AnotherVertex(nowVertex)];
	  vec[1]=vertex[nowVertex]-vertex[edge[startEdge[1]].
					 AnotherVertex(nowVertex)];

	  tmpN2=vec[0].Cross(vec[1]);
	  tmpN2.Normalize();
#ifdef DEBUG
	  cout << "base normal==";
	  tmpN2.Print();
#endif
	  if(isnan(tmpN2.coord[0])){
	    DebugOut("prevent nan : set zero");
	    tmpN.Set(0,0,0);
	  }
	  totalNormal.Plus(tmpN2); // add this face's normal
	  faceSum++;
	  for(int cnt3=0;cnt3<startEdge.size();cnt3++){
	    DebugOut("searching next face from edge[" << startEdge[cnt3] << "]...");
	    int oldEdge=startEdge[cnt3],newEdge;
	    int oldFace=nowFace,newFace;	    
	    vec[0]=vertex[nowVertex]-vertex[edge[startEdge[0]].
					   AnotherVertex(nowVertex)];
	    vec[1]=vertex[nowVertex]-vertex[edge[startEdge[1]].
					   AnotherVertex(nowVertex)];
	    if(cnt3!=0){
	      swap(vec[0],vec[1]);
	    }
	    int whileFlag=1;
	    while(edge[oldEdge].facePLink.size()==2 && whileFlag){
	      DebugOut("found new face.");
	      newFace=edge[oldEdge].AnotherFace(oldFace); // newFace
	      DebugOut("oldEdge,oldFace,newFace==" << oldEdge << ", "
		       << oldFace << ", " << newFace);
	      if(newFace==nowFace){ // prevent loop
		DebugOut("but newFace is nowFace. so end nowFace.");
		cnt3=startEdge.size();
		whileFlag=0;
		continue;
	      }
	      vec[1]=vec[0];
	      
	      int tmpEdge;
	      for(int cnt4=0;cnt4<nowVertexP->edgePLink.size();cnt4++){
		tmpEdge=nowVertexP->edgePLink[cnt4];
		if(face[newFace].HaveEdge(tmpEdge) && tmpEdge!=oldEdge){
		  newEdge=tmpEdge; // newEdge
		  cnt4=nowVertexP->edgePLink.size(); // exit this for-loop
		}
	      }
	      DebugOut("newEdge==" << newEdge);
	      vec[0]=vertex[nowVertex]
		-vertex[edge[newEdge].AnotherVertex(nowVertex)]; // another vector
	      if(face[newFace].HasBreak(nowVertex)){
		DebugOut("nowVertex in newFace has break");
		vec[2]=vertex[nowVertex]
		  -vertex[face[newFace].OppositeVertex(nowVertex)];
		Vector3d tmpVec;
		tmpVec.Cross(vec[0],vec[2]);
		tmpVec.Normalize();
		if(isnan(tmpVec.coord[0])){
		  DebugOut("prevent nan : through this face");
		  tmpVec.Set(0,0,0); // prevent nan
		}
		tmpN.Cross(vec[2],vec[1]);
		tmpN.Normalize();
		if(isnan(tmpN.coord[0])){
		  DebugOut("prevent nan : set zero2");
		  tmpN.Set(0,0,0); // prevent nan
		}
		tmpN.Plus(tmpVec);
		tmpN.Normalize();
	      }else{
		tmpN.Cross(vec[0],vec[1]);
		tmpN.Normalize();
	      }
	      if(isnan(tmpN.coord[0])){
		DebugOut("prevent nan:abort this face2 -> set dummy");
		tmpN.Set(0,0,0);
	      }
	      
	      if(cnt3==0){
		totalNormal.Plus(tmpN); // add this face's normal
	      }else{
		totalNormal.Minus(tmpN); // add this face's normal
	      }
	      faceSum++;
#ifdef DEBUG
	      cout << "now totalNormal==";
	      totalNormal.Print();
#endif
	      oldFace=newFace; // store face number
	      oldEdge=newEdge; // store edge number
	    }
	    DebugOut("end search.");
	  }
	  totalNormal.Normalize();
	  if(isnan(totalNormal.coord[0])){
	    DebugOut("set totalNormal 0,0,0");
	    totalNormal.Set(0,0,0);
	  }
#ifdef DEBUG
	  printf("\tfaceSum==%d\n",faceSum);
	  printf("\tAverage normal==");
	  totalNormal.Print();
#endif
	  fprintf(fp,"\t\t<%f, %f, %f>,<%f, %f, %f>",
		  vertex[nowVertex].X(),
		  vertex[nowVertex].Y(),
		  -vertex[nowVertex].Z(),
		  totalNormal.X(),
		  totalNormal.Y(),
		  -totalNormal.Z()); 	  
	  if(cnt2!=2){
	    fprintf(fp,",\n");
	  }else{
	    fprintf(fp,"}\n");
	  }
	}

      }
    }
    break;
  }

  fprintf(fp,"\ttexture {\n"); // texture
  fprintf(fp,"\t\tpigment {\n"); // pigment
  fprintf(fp,"\t\t\t rgb <1,1,1>\n");
  fprintf(fp,"\t\t}\n"); // pigment
  fprintf(fp,"\t\tfinish {\n"); // finish
  fprintf(fp,"\t\t\tdiffuse 0.5\n");
  fprintf(fp,"\t\t\tambient 0.1\n");
  fprintf(fp,"\t\t\tspecular .5 roughness .05\n");
  fprintf(fp,"\t\t}\n"); // finish
  fprintf(fp,"\t}\n"); // texture
  fprintf(fp,"}\n"); // union

  fclose(fp);
  for(cnt1=0;cnt1<dif;cnt1++){
    Subdivide(-1);
#ifdef DEBUG
    cout << "unsub..." << endl;
#endif
  }

  endTime=time(NULL);
  workTime=difftime(endTime,startTime);  
#ifdef DEBUG
  cout << "Export time : " << workTime << endl;
  cout << "Exported." << endl;
#endif
}

int
GetString(char* exportType,char* defaultSub){
#ifdef DEBUG
  cout << defaultSub << endl;
#endif
  int mouseX=Fl::event_x_root();
  int mouseY=Fl::event_y_root();
  Fl_Window povExportW(mouseX,mouseY,225,65+30,exportType);
  Fl_Int_Input povInput(120,5,100,20,"Subdivision Level");
  povInput.value(defaultSub);
  Fl_Choice camChoice(120,28,90,30,"select camera");
  char camIndex[1024];
  sprintf(camIndex,"");
  for(int cnt1=0;cnt1<cameraList.size();cnt1++){
    strcat(camIndex,cameraList[cnt1].name);
    strcat(camIndex," - ");
    switch(cameraList[cnt1].type){
      //case NORMAL:
    case PERSPECTIVE:
      strcat(camIndex,"Perspective");
      break;
      //case XY:
    case XYMP:
      strcat(camIndex,"XY-Ortho:Z ->+");
      break;
    case XYPM:
      strcat(camIndex,"XY-Ortho:Z +>-");
      break;
      //case YZ:
    case YZMP:
      strcat(camIndex,"YZ-Ortho:X ->+");
      break;
    case YZPM:
      strcat(camIndex,"YZ-Ortho:X +>-");
      break;
      //case XZ:
    case XZMP:
      strcat(camIndex,"XZ-Ortho:Y ->+");
      break;      
    case XZPM:
      strcat(camIndex,"XZ-Ortho:Y +>-");
      break;      
    }
    strcat(camIndex," - ");
    char widthT[10];
    char heightT[10];
    sprintf(widthT,"%dx",cameraList[cnt1].width);
    sprintf(heightT,"%d",cameraList[cnt1].height);
    strcat(camIndex,widthT);
    strcat(camIndex,heightT);
    strcat(camIndex," - ");
    char camOrder[20];
    sprintf(camOrder,"camera[%d]",cnt1);
    strcat(camIndex,camOrder);
    strcat(camIndex,"|");
  }
  camChoice.add(camIndex);
  camChoice.value(0);
  Fl_Return_Button ok(30,30+30,70,30,"ok");
  Fl_Button cancel(120,30+30,70,30,"cancel");
  povExportW.hotspot(&cancel);
  //cancel.callback((Fl_Callback*)CancelCB);
  //ok.callback((Fl_Callback*)ExportPOVCore);
  
  povExportW.end();
  povExportW.set_modal();
  povExportW.show();
  for (;;) {
    Fl::wait();
    Fl_Widget *o;
    while ((o = Fl::readqueue())) {
      if (o == &ok) {
        strcpy(defaultSub,povInput.value());
	outputC=camChoice.value();
        return atoi(povInput.value());
      } else if (o == &cancel || o == &povExportW) {
        return -1;
      }
    }
  }
}

void
ExportOFF(){
  DebugOut("ExportOFF");
  char offFile[255];
  char tmp[255];
  int cnt1;

  time_t startTime;
  time_t endTime;
  double workTime;
  // Time 
  startTime=time(NULL);

  int level=GetString("OFF Export",defaultSub);
  if(level<0){
    return;
  }
  DebugOut("In ExportOFF;filename==" << filename);
  if(!strcmp(filename,"")){
    sprintf(offFile,"Untitled.off");
  }else{
    
    DebugOut("strlen==" << strlen(filename));
    for(cnt1=0;cnt1<strlen(filename)-4;cnt1++){
      tmp[cnt1]=filename[cnt1];
    }
    tmp[cnt1]='\0';
    DebugOut("tmp==" << tmp);
    //strncpy(tmp,filename,(int)strlen(filename)-4);
    sprintf(offFile,"%s.off",tmp);
  }

  FILE* fp;
  int flag;
  fp=fopen(offFile,"r"); 
  if(fp!=NULL){
    char mes[255];
    sprintf(mes,"%s already exist. Overwrite?",offFile);
    flag=fl_choice(mes,"yes","no",NULL);
    if(flag==1){ // no
#ifdef DEBUG
      cout << "In ExportOFF;return" << endl;
#endif
      return;
    }else { // 2==yes
      fclose(fp);
      fp=fopen(offFile,"w");
#ifdef DEBUG
      cout << "Willing write..." << endl;
#endif
    }
  }else{
    fp=fopen(offFile,"w");
#ifdef DEBUG
    cout << "Willing write also..." << endl;
#endif
  }
  
  int dif=level-subdivideLevel;
  for(cnt1=0;cnt1<dif;cnt1++){
    Subdivide(1);
  }
  
  // file output
  fprintf(fp,"OFF\n");
  fprintf(fp,"%d %d %d\n",
	  subdividedVertex[level+1]-subdividedVertex[level],
	  subdividedFace[level+1]-subdividedFace[level],
	  subdividedEdge[level+1]-subdividedEdge[level]);
  for(int cnt1=subdividedVertex[level];cnt1<subdividedVertex[level+1];cnt1++){
    fprintf(fp,"%f %f %f\n",
	    vertex[cnt1].X(),
	    vertex[cnt1].Y(),
	    vertex[cnt1].Z());
  }
  for(int cnt1=subdividedFace[level];cnt1<subdividedFace[level+1];cnt1++){
    if(face[cnt1].vertexLink.size()==3){
      fprintf(fp,"%d %d %d %d\n",3,
	      face[cnt1].vertexLink[0]-subdividedVertex[level],
	      face[cnt1].vertexLink[1]-subdividedVertex[level],
	      face[cnt1].vertexLink[2]-subdividedVertex[level]);
    }else {
      fprintf(fp,"%d %d %d %d %d\n",4,
	      face[cnt1].vertexLink[0]-subdividedVertex[level],
	      face[cnt1].vertexLink[1]-subdividedVertex[level],
	      face[cnt1].vertexLink[2]-subdividedVertex[level],
	      face[cnt1].vertexLink[3]-subdividedVertex[level]);
    }
  }  
  fclose(fp);
}

// triangle
void
ExportPOV(){
  DebugOut("POV-Ray triangle Export");

  int level=GetString("POV-Ray Export:triangle",defaultSub);
  if(level>=0){
    ExportPOVCore(level,TRIANGLE);
  }
  
  DebugOut(defaultSub);
}


void
ExportPOVSmooth(){
  DebugOut("POV-Ray smooth_triangle Export");


  int level=GetString("POV-Ray Export:smooth_triangle",defaultSub);
  if(level>=0){
    ExportPOVCore(level,SMOOTH_TRIANGLE);
  }
  DebugOut(defaultSub);
}


void
ImportDXF(){
  
}

void
ExportDXF(){
  DebugOut("ExportDXF");
  char dxfFile[255];
  char tmp[255];
  int cnt1;

  time_t startTime;
  time_t endTime;
  double workTime;
  // Time 
  startTime=time(NULL);

  int level=GetString("DXF Export",defaultSub);
  if(level<0){
    return;
  }
  DebugOut("In ExportDXF;filename==" << filename);
  if(!strcmp(filename,"")){
    sprintf(dxfFile,"Untitled.dxf");
  }else{
    
    DebugOut("strlen==" << strlen(filename));
    for(cnt1=0;cnt1<strlen(filename)-4;cnt1++){
      tmp[cnt1]=filename[cnt1];
    }
    tmp[cnt1]='\0';
    DebugOut("tmp==" << tmp);
    //strncpy(tmp,filename,(int)strlen(filename)-4);
    sprintf(dxfFile,"%s.dxf",tmp);
  }

  FILE* fp;
  int flag;
  fp=fopen(dxfFile,"r"); 
  if(fp!=NULL){
    char mes[255];
    sprintf(mes,"%s already exist. Overwrite?",dxfFile);
    flag=fl_choice(mes,"yes","no",NULL);
    if(flag==1){ // no
#ifdef DEBUG
      cout << "In ExportDXF;return" << endl;
#endif
      return;
    }else { // 2==yes
      fclose(fp);
      fp=fopen(dxfFile,"w");
#ifdef DEBUG
      cout << "Willing write..." << endl;
#endif
    }
  }else{
    fp=fopen(dxfFile,"w");
#ifdef DEBUG
    cout << "Willing write also..." << endl;
#endif
  }
  
  int dif=level-subdivideLevel;
  for(cnt1=0;cnt1<dif;cnt1++){
    Subdivide(1);
  }
  
  // file output
  fprintf(fp,"  0\n");
  fprintf(fp,"SECTION\n");
  fprintf(fp,"  2\n");
  fprintf(fp,"BLOCKS\n");
  fprintf(fp,"  0\n");
  fprintf(fp,"ENDSEC\n");
  fprintf(fp,"  0\n");
  fprintf(fp,"SECTION\n");
  fprintf(fp,"  2\n");
  fprintf(fp,"ENTITIES\n");
  fprintf(fp,"  0\n");
  fprintf(fp,"POLYLINE\n");
  fprintf(fp,"  8\n");
  // point,face,number
  fprintf(fp,"0\n");
  fprintf(fp," 62\n");
  fprintf(fp,"\t1\n");
  fprintf(fp," 66\n");
  fprintf(fp,"\t1\n");
  fprintf(fp," 10\n");
  fprintf(fp,"0.0\n");
  fprintf(fp," 20\n");
  fprintf(fp,"0.0\n");
  fprintf(fp," 30\n");
  fprintf(fp,"0.0\n");
  fprintf(fp," 70\n");
  fprintf(fp,"\t64\n");
  fprintf(fp," 71\n");
  fprintf(fp,"\t%d\n",subdividedVertex[level+1]-subdividedVertex[level]);
  fprintf(fp," 72\n");
  fprintf(fp,"\t%d\n",subdividedFace[level+1]-subdividedFace[level]);
  fprintf(fp," 0\n"); 
  
  // eace point
  for(int cnt1=subdividedVertex[level];cnt1<subdividedVertex[level+1];cnt1++){
    fprintf(fp,"VERTEX\n");
    fprintf(fp,"  8\n");
    fprintf(fp,"0\n");
    fprintf(fp," 62\n");
    fprintf(fp,"\t1\n"); // layer number
    fprintf(fp," 10\n"); // x
    fprintf(fp,"%f\n",vertex[cnt1].X());
    fprintf(fp," 20\n"); // y
      fprintf(fp,"%f\n",vertex[cnt1].Y());
      fprintf(fp," 30\n"); // z
      fprintf(fp,"%f\n",vertex[cnt1].Z());
      fprintf(fp," 70\n");
      fprintf(fp,"\t192\n");
      fprintf(fp," 0\n");
  }
  // eace face
  for(int cnt1=subdividedFace[level];cnt1<subdividedFace[level+1];cnt1++){
    fprintf(fp,"VERTEX\n");
      fprintf(fp,"  8\n");
      fprintf(fp,"0\n");
      fprintf(fp," 62\n");
      fprintf(fp,"\t%d\n",face[cnt1].layer); // layer number
      fprintf(fp," 10\n"); // x
      fprintf(fp,"0.0\n");
      fprintf(fp," 20\n"); // y
      fprintf(fp,"0.0\n");
      fprintf(fp," 30\n"); // z
      fprintf(fp,"0.0\n");
      fprintf(fp," 70\n");
      fprintf(fp,"\t128\n");
      fprintf(fp," 71\n");      
      fprintf(fp,"\t%d\n",face[cnt1].vertexLink[0]-subdividedVertex[level]+1);
      fprintf(fp," 72\n");      
      fprintf(fp,"\t%d\n",face[cnt1].vertexLink[1]-subdividedVertex[level]+1);
      fprintf(fp," 73\n");      
      fprintf(fp,"\t%d\n",face[cnt1].vertexLink[2]-subdividedVertex[level]+1);
      if(face[cnt1].vertexLink.size()==4){
	  fprintf(fp," 74\n");      
	  fprintf(fp,"\t%d\n",face[cnt1].vertexLink[3]-subdividedVertex[level]+1);
      }           
      fprintf(fp," 0\n");
  }
  fprintf(fp,"SEQEND\n");
  fprintf(fp,"\t8\n");
  fprintf(fp,"0\n");
  fprintf(fp," 62\n");
  fprintf(fp,"\t7\n");
  fprintf(fp,"  0\n");
  fprintf(fp,"ENDSEC\n");
  fprintf(fp,"  0\n");
  fprintf(fp,"EOF\n");
  fclose(fp);
}



